@@ -4,7 +4,7 @@ use Carp ();
 ############################################################################
 package Net::OpenID::ClaimedIdentity;
 BEGIN {
-  $Net::OpenID::ClaimedIdentity::VERSION = '1.030099_005';
+  $Net::OpenID::ClaimedIdentity::VERSION = '1.030099_006';
 }
 use fields (
     'identity',         # the canonical URL that was found, following redirects
@@ -111,6 +111,7 @@ sub check_url {
         $assoc = Net::OpenID::Association::handle_assoc($csr, $ident_server, $use_assoc_handle);
     } else {
         $assoc = Net::OpenID::Association::server_assoc($csr, $ident_server, $force_reassociate, (
+            %{$csr->assoc_options},
             protocol_version => $self->protocol_version,
         ));
     }
@@ -178,19 +179,18 @@ sub check_url {
     foreach my $ext_uri (keys %{$self->{extension_args}}) {
         my $ext_alias;
 
-        my $version=$self->protocol_version;
-        $version=$1 if $ext_uri=~m!/(\d+(:?[.]\d+))$!;
-        if ($version >= 2) {
-            $ext_alias = 'e'.($ext_idx++);
-            $ext_url_args{'openid.ns.'.$ext_alias} = $ext_uri;
-        }
-        else {
+        if ($ext_uri eq "http://openid.net/extensions/sreg/1.1") {
             # For OpenID 1.1 only the "SREG" extension is allowed,
             # and it must use the "openid.sreg." prefix.
-            next unless $ext_uri eq "http://openid.net/extensions/sreg/1.1";
             $ext_alias = "sreg";
-            $ext_url_args{'openid.ns.sreg'}=$ext_uri;
         }
+        elsif ($self->protocol_version < 2) {
+            next;
+        }
+        else {
+            $ext_alias = 'e'.($ext_idx++);
+        }
+        $ext_url_args{'openid.ns.'.$ext_alias} = $ext_uri;
 
         foreach my $k (keys %{$self->{extension_args}{$ext_uri}}) {
             $ext_url_args{'openid.'.$ext_alias.'.'.$k} = $self->{extension_args}{$ext_uri}{$k};
@@ -213,7 +213,7 @@ Net::OpenID::ClaimedIdentity - A not-yet-verified OpenID identity
 
 =head1 VERSION
 
-version 1.030099_005
+version 1.030099_006
 
 =head1 SYNOPSIS
 
